//VP FCUP 2022-2023
datatype Nat = Z | S(Nat)

function   Plus(x:Nat, y:Nat): Nat
decreases x;
{
match x {
    case Z => y
    case S(z) => S(Plus(z,y))
}
}

lemma {:induction n} example(n:Nat)
decreases n;
ensures Plus(n,Z) == n
{}

lemma {:induction false} example0 (n: Nat)
decreases n
ensures Plus(n,Z) == n
{ 
    match n {
    case Z => { assert Plus(Z,Z) == Z;}
    case S(k) => {
        assert Plus(n,Z) == S(Plus(k,Z));
        example0(k);
        assert S(Plus(k,Z)) == S(k);
        assert  S(k) == n;
        }
    }

}

function UnaryToNat(x: Nat): nat {
 match x
  case Z => 0
  case S(y) => 1 + UnaryToNat(y)
}

function NatToUnary(n: nat): Nat {
 if n == 0 then Z else S(NatToUnary(n-1))
}

lemma NatUnaryCorrespondence(n: nat, x: Nat)
 ensures UnaryToNat(NatToUnary(n)) == n
 ensures NatToUnary(UnaryToNat(x)) == x
{
}

lemma  PlusCorrect(x: Nat, y: Nat)
 ensures UnaryToNat(Plus(x, y)) == UnaryToNat(x) + UnaryToNat(y)
{
}

lemma {:induction false} PlusICorrect(x: Nat, y: Nat)
  ensures UnaryToNat(Plus(x, y)) == UnaryToNat(x) + UnaryToNat(y)
{
  match x {
  case Z =>
  case S(x') =>
    calc {
      UnaryToNat(Plus(x, y));
    ==  // y == Suc(y')
      UnaryToNat(Plus(S(x'),y));
    ==  // def. Plus
      UnaryToNat(S(Plus(x', y)));
    ==  // def. UnaryToNat
      1 + UnaryToNat(Plus(x', y));
    ==  { PlusICorrect(x', y); }
      1 + UnaryToNat(x') + UnaryToNat(y);
    ==
      UnaryToNat(x')  + 1 + UnaryToNat(y);
    ==  // def. UnaryToNat
      UnaryToNat(S(x')) + UnaryToNat(y);
    ==  // y == S(y')
      UnaryToNat(x) + UnaryToNat(y);
    }
  }
}

function Mul(x: Nat, y: Nat): Nat {
 match x
 case Z => Z
 case S(x') => Plus(Mul(x', y), y)
}


lemma MulCorrect(x: Nat, y: Nat)
 ensures UnaryToNat(Mul(x, y)) == UnaryToNat(x) * UnaryToNat(y)
{
 match x
 case Z =>
 case S(x') =>
 calc {
   UnaryToNat(Mul(x, y));
  ==  // def. Mul
   UnaryToNat(Plus(Mul(x', y), y));
  ==  { PlusCorrect(Mul(x', y), y); }
   UnaryToNat(Mul(x', y)) + UnaryToNat(y);

 }
}
