
type T = int
class Stack
{
    var elems: array<T>;
    var size : nat;
    var capacity: nat;

    ghost predicate Valid()
      reads this
    {
        size <= capacity == elems.Length
    }
 
    constructor (cap: nat)
      requires cap > 0
      ensures fresh(elems)
      ensures Valid() && capacity == cap && size == 0
    {
        elems := new T[cap];
        capacity := cap;
        size := 0; 
    }

  method isEmpty() returns(res: bool)
      requires Valid()
      ensures res == (size == 0)
    {
        return size == 0;
    }

    method  isFull() returns (res:bool)
    requires Valid()
    ensures res == (size == elems.Length)
    {
        return size == elems.Length;
    }
    method push(x : T)
      modifies  this, elems
      requires Valid() && (size < capacity)
      ensures Valid()
      ensures elems[..size] == old(elems[..size]) + [x]
    {
        elems[size] := x;
        size := size + 1;
    }

     method top() returns (x: T)
      requires Valid() && size > 0
      ensures x == elems[size-1]
    {
        return elems[size-1];
    }
    
    method pop() 
      requires Valid() && size > 0
      modifies this
      ensures Valid() 
      ensures elems[..size] == old(elems[..size-1])
    {
         size := size-1;
    }
}

/* // A simple test case to check on the console.
method {:verify false} Main()
{
    var s := new Stack(3);
    s.push(1);
    s.isEmpty();
    s.push(2);
    s.push(3);
    s.pop();
    var x := s.top();
    print "top=", x, " (should be 2)\n";
} */