type T = int 
 
class Stack
{
    const elems: array<T>; 
    var size : nat; 
    var capacity: nat;
		
    constructor (cap: nat)
    requires cap > 0
    ensures fresh(elems)
    ensures Valid() && cap == capacity && size ==0
    {
        elems := new T[cap];
        capacity := cap;
        size := 0; 
    }
    predicate Valid()
    reads this
    { size <= capacity == elems.Length
    }
		
    predicate  isEmpty() 
    reads this
    {
        size == 0
    }
 
    ghost predicate   isFull()
			requires Valid()
			
			reads this,elems
			
    ensures isFull() <==> size == capacity
    {
       size == elems.Length
    }
 
    method  push(x : T)
    modifies this, elems
    requires Valid() &&  ! isFull() //size < capacity
    ensures Valid()
    ensures elems[..size] == old(elems[..size]) + [x]
    {
        elems[size] := x;
        size := size + 1;
    }
 
    function  top(): T
    requires Valid() && size > 0
    reads this, elems 
    {
         elems[size-1]
    }
    
    method pop() 
    requires Valid() && size > 0
    modifies this
    ensures Valid() 
    ensures elems[..size] == old(elems[..size-1])
    {
         size := size-1;
    }
}
 
// A simple test case.
method {:verify true}  Main()
{
    var s := new Stack(3);
    assert s.isEmpty();
    s.push(1);
    s.push(2);
    s.push(3);
    assert s.top() == 3;
    assert s.isFull();
    s.pop();
    assert s.top() == 2;
    print "top = ", s.top(), " \n";
}
